// iphlpadv.h

/*
_MSC_VER
 Reports the major and minor versions of the compiler. 
 For example, 1310 for Microsoft Visual C++ .NET 2003. 
 1310 represents version 13 and a 1.0 point release. 
 The current compiler version is 1400.
*/
#if _MSC_VER < 1400
typedef struct _MIB_UDPROW_OWNER_PID {
    DWORD dwLocalAddr;
    DWORD dwLocalPort;
    DWORD dwOwningPid;
} MIB_UDPROW_OWNER_PID, *PMIB_UDPROW_OWNER_PID;
typedef struct _MIB_UDPTABLE_OWNER_PID
{
    DWORD                   dwNumEntries;
    MIB_UDPROW_OWNER_PID    table[ANY_SIZE];
} MIB_UDPTABLE_OWNER_PID, *PMIB_UDPTABLE_OWNER_PID;
#endif

typedef enum  {
  UDP_TABLE_BASIC,
  UDP_TABLE_OWNER_PID,
  UDP_TABLE_OWNER_MODULE 
} UDP_TABLE_CLASS, *PUDP_TABLE_CLASS;

typedef DWORD (WINAPI *pfnGetExtendedUdpTable)(PVOID, PDWORD, BOOL, ULONG, UDP_TABLE_CLASS, ULONG);

static SERVICE_FAILURE_ACTIONS asfa[1024*8/sizeof(SERVICE_CONFIG_FAILURE_ACTIONS)+1]; // max size 1024*8 bytes
static BOOL service_config_changed(FALSE);

DWORD SearchPidPortListenerUDP(DWORD port)
{
	MIB_UDPTABLE_OWNER_PID *pts(NULL);
	DWORD stsize(0), res(0);

	HMODULE hModule = ::GetModuleHandle(TEXT("Iphlpapi.dll"));
	if(!hModule) {
		hModule = ::LoadLibrary(TEXT("Iphlpapi.dll"));
		if(!hModule) return 0;
	}
	pfnGetExtendedUdpTable GetExtendedUdpTable = (pfnGetExtendedUdpTable) ::GetProcAddress(hModule, "GetExtendedUdpTable"); 
	res = GetExtendedUdpTable(pts, &stsize, TRUE, AF_INET, UDP_TABLE_OWNER_PID, 0L);
	if(ERROR_INSUFFICIENT_BUFFER == res) {
		pts = new MIB_UDPTABLE_OWNER_PID[stsize/sizeof(MIB_UDPTABLE_OWNER_PID)];
		res = GetExtendedUdpTable(pts, &stsize, TRUE, AF_INET, UDP_TABLE_OWNER_PID, 0L);
	}
	res = 0;
	int count = pts[0].dwNumEntries;
	for(int i=0; i < count; i++) {
		if(pts->table[i].dwLocalPort == port) {
			res = pts->table[i].dwOwningPid;
			break;
		}
	}
	if(stsize) delete [] pts;
	::FreeLibrary(hModule);
	return res;
}
BOOL KillProcessByID(DWORD pid)
{
	HANDLE h = ::OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_TERMINATE, TRUE, pid);
	if(!h) return FALSE;
	BOOL t = ::TerminateProcess(h, 0);
	::CloseHandle(h);
	return t;
}
BOOL CheckRestartService(SC_HANDLE h)
{
	BYTE buffer[1024*8];
	DWORD dwBytesNeeded(0);
	BOOL t = ::QueryServiceConfig2(h, SERVICE_CONFIG_FAILURE_ACTIONS, buffer, sizeof(buffer), &dwBytesNeeded);
	SERVICE_FAILURE_ACTIONS &sfa = *(SERVICE_FAILURE_ACTIONS*)buffer;
	for(unsigned int i=0; i<sfa.cActions; i++) {
		if(sfa.lpsaActions[i].Type == SC_ACTION_RESTART ||
			sfa.lpsaActions[i].Type == SC_ACTION_REBOOT) return TRUE;
	}
	return FALSE;
}
void ChangeServiceConfigSetRestartNone(SC_HANDLE h)
{
	BYTE buffer[1024*8];
	DWORD dwBytesNeeded(0);
	BOOL t = ::QueryServiceConfig2(h, SERVICE_CONFIG_FAILURE_ACTIONS, buffer, sizeof(buffer), &dwBytesNeeded);
	memcpy(&asfa, buffer, sizeof(buffer));
	SERVICE_FAILURE_ACTIONS &sfa = *(SERVICE_FAILURE_ACTIONS*)buffer;
	sfa.cActions = 0;
	sfa.lpsaActions = NULL;
	t = ChangeServiceConfig2(h, SERVICE_CONFIG_FAILURE_ACTIONS, &sfa);
	return;
}
TCHAR* StopServiceByPID(DWORD pid)
{
	BOOL t(FALSE);
	DWORD BytesNeeded(0), ServicesReturned(0), ResumeHandle(0);
	static TCHAR name[256] = { 0 };
	SERVICE_STATUS ss;
	ENUM_SERVICE_STATUS_PROCESS* pinfo(NULL);
	SC_HANDLE hscm(0), hsvc(0);
	BYTE* ptr(0);

	hscm = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	t = ::EnumServicesStatusEx(hscm, SC_ENUM_PROCESS_INFO, SERVICE_WIN32, SERVICE_STATE_ALL, 
		(LPBYTE)pinfo,
		0,
		&BytesNeeded,
		&ServicesReturned,
		&ResumeHandle,
		NULL);
	ptr = new BYTE[BytesNeeded];
	pinfo = (ENUM_SERVICE_STATUS_PROCESS*)ptr;
	t = ::EnumServicesStatusEx(hscm, SC_ENUM_PROCESS_INFO, SERVICE_WIN32, SERVICE_STATE_ALL, 
		(LPBYTE)pinfo,
		BytesNeeded,
		&BytesNeeded,
		&ServicesReturned,
		&ResumeHandle,
		NULL);
	for(unsigned int i=0; i<ServicesReturned; i++) {
		if(pinfo[i].ServiceStatusProcess.dwProcessId == pid) {
			hsvc = ::OpenService(hscm, pinfo[i].lpServiceName, SC_MANAGER_ALL_ACCESS);
			t = ::ControlService(hsvc, SERVICE_CONTROL_STOP, &ss);
			if(t) {
				lstrcpy(name, pinfo[i].lpServiceName);
				if(CheckRestartService(hsvc)) {
					SC_LOCK sclLock = LockServiceDatabase(hscm);
					ChangeServiceConfigSetRestartNone(hsvc);
					service_config_changed = TRUE;
					UnlockServiceDatabase(sclLock);
				}
			}
			break;
		}
	}
	::CloseServiceHandle(hsvc);
	::CloseServiceHandle(hscm);
	delete [] ptr;
	return name;
}
void StartServiceByName(TCHAR* name)
{
	BOOL t(FALSE);
	SC_HANDLE hscm(0), hsvc(0);

	hscm = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	hsvc = ::OpenService(hscm, name, SERVICE_START);
	if(service_config_changed) {
		SC_LOCK sclLock = LockServiceDatabase(hscm);
		t = ChangeServiceConfig2(hsvc, SERVICE_CONFIG_FAILURE_ACTIONS, asfa);
		UnlockServiceDatabase(sclLock);
		service_config_changed = FALSE;
	}
	t = ::StartService(hsvc, 0, NULL);
	::CloseServiceHandle(hsvc);
	::CloseServiceHandle(hscm);
	return;
}
