// syslog.cpp : Defines the entry points for the DLL application.

#include "stdafx.h"
#include "syslog514.h"
#include "eventlog.h"
/*
	Export forward declaration
*/
extern "C" __declspec(dllexport) void ServiceMain(DWORD argc, LPTSTR *argv);
extern "C" __declspec(dllexport) HRESULT DllRegisterServer(void);
extern "C" __declspec(dllexport) HRESULT DllUnregisterServer(void);
void WINAPI CtrlHandler(DWORD opcode); 
DWORD WINAPI Initialization(DWORD argc, LPTSTR *argv, DWORD *specificError);
/*
	Shared module data section
*/
#pragma data_seg(".shared")
WORD	wLoadCounter(0);
TCHAR	czGModuleFilename[MAX_PATH] = { 0 };
CSysLog	*psyslog(NULL);
#pragma comment(linker, "/SECTION:.shared,RWS")
#pragma data_seg()
/*
	Service members
*/
SERVICE_STATUS          ServiceStatus; 
SERVICE_STATUS_HANDLE   ServiceStatusHandle; 
/*
	Debug routine
*/
#if defined(_DEBUG)
void MyOutputDebug(LPVOID s, WORD size = 0)
{
	DWORD dwWritten(0);
	HANDLE h = ::CreateFile(_T("myoutput.txt"), FILE_WRITE_DATA, FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	::SetFilePointer(h, NULL, NULL, FILE_END);
	if(!size) {
		TCHAR* p = (TCHAR*)s;
		while(*p) p++;
		::WriteFile(h, s, p-s, &dwWritten, NULL);
	} else {
		::WriteFile(h, s, size, &dwWritten, NULL);
	}
	::CloseHandle(h);
	return;
}
#else
#define MyOutputDebug //
#endif
/*
	Every load & call entry point. Main entry point. Very main. Very-very main.
*/
BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	switch(ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		MyOutputDebug(_T("DllMain:DLL_PROCESS_ATTACH\n"));
		::DisableThreadLibraryCalls((HMODULE)hModule);
		MyOutputDebug(::GetCommandLine());
		MyOutputDebug(_T("\n"));
		if(!wLoadCounter) {
			psyslog = new CSysLog;
			::GetModuleFileName((HMODULE)hModule, czGModuleFilename, MAX_PATH);
			MyOutputDebug(&wLoadCounter, sizeof(wLoadCounter));
			MyOutputDebug(_T("\n"));
		}
		wLoadCounter++;
		MyOutputDebug(czGModuleFilename);
		MyOutputDebug(_T("\n"));
		break;
	case DLL_THREAD_ATTACH:
		MyOutputDebug(_T("DllMain:DLL_THREAD_ATTACH\n"));
		break;
	case DLL_THREAD_DETACH:
		MyOutputDebug(_T("DllMain:DLL_THREAD_DETACH\n"));
		break;
	case DLL_PROCESS_DETACH:
		MyOutputDebug(_T("DllMain:DLL_PROCESS_DETACH\n\n"));
		wLoadCounter--;
		if(!wLoadCounter) {
			MyOutputDebug(_T("delete psyslog\n"));
			delete psyslog;
		}
		MyOutputDebug(&wLoadCounter, sizeof(wLoadCounter));
		MyOutputDebug(_T("\n"));
		break;
	default:
		MyOutputDebug(_T("DllMain:UNDEFINED\n"));
	}
	return TRUE;
}
void ServiceMain(DWORD argc, LPTSTR *argv)
{
	MyOutputDebug(_T("ServiceMain\n"));
	CSysLog	&syslog = *psyslog;
	//::DebugBreak();
    DWORD status; 
    ServiceStatus.dwServiceType        = SERVICE_WIN32_OWN_PROCESS; 
    ServiceStatus.dwCurrentState       = SERVICE_START_PENDING; 
    ServiceStatus.dwControlsAccepted   = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_PAUSE_CONTINUE; 
    ServiceStatus.dwWin32ExitCode      = 0; 
    ServiceStatus.dwServiceSpecificExitCode = 0; 
    ServiceStatus.dwCheckPoint         = 0; 
    ServiceStatus.dwWaitHint           = 1000; 
	ServiceStatusHandle = ::RegisterServiceCtrlHandler(_T("SysLogService"), CtrlHandler); 
    if(ServiceStatusHandle == (SERVICE_STATUS_HANDLE)0) { 
		MyOutputDebug(_T("ServiceStatusHandle return\n"));
        return; 
    } 
	syslog.Enjoy();
    ServiceStatus.dwCurrentState       = SERVICE_RUNNING; 
    ServiceStatus.dwCheckPoint         = 0; 
    ServiceStatus.dwWaitHint           = 0; 
 	if(!::SetServiceStatus(ServiceStatusHandle, &ServiceStatus)) { 
		MyOutputDebug(_T("SetServiceStatus ServiceMain()\n"));
		status = ::GetLastError(); 
    }
    return; 
} 
void WINAPI CtrlHandler(DWORD Opcode) 
{ 
	MyOutputDebug(_T("CtrlHandler\n"));
	DWORD status; 
	CSysLog	&syslog = *psyslog;
	switch(Opcode) { 
		case SERVICE_CONTROL_PAUSE: 
			syslog.Pause();
			ServiceStatus.dwCheckPoint++; 
			ServiceStatus.dwCurrentState = SERVICE_PAUSED; 
			break; 
		case SERVICE_CONTROL_CONTINUE: 
			syslog.Resume();
			ServiceStatus.dwCheckPoint++; 
			ServiceStatus.dwCurrentState = SERVICE_RUNNING; 
			break; 
		case SERVICE_CONTROL_STOP: 
			ServiceStatus.dwCurrentState  = SERVICE_STOP_PENDING;
			ServiceStatus.dwCheckPoint++; 
			ServiceStatus.dwWaitHint      = 500; 
			if(!::SetServiceStatus(ServiceStatusHandle,  &ServiceStatus)) { 
				status = ::GetLastError(); 
			}
			syslog.Shutdown();
			while(CSysLog::esStopped != syslog.State()) ::Sleep(50);
			ServiceStatus.dwWin32ExitCode = 0; 
			ServiceStatus.dwCurrentState  = SERVICE_STOPPED; 
			ServiceStatus.dwCheckPoint++; 
			ServiceStatus.dwWaitHint      = 100; 
			break;
		case SERVICE_CONTROL_INTERROGATE: 
			MyOutputDebug(_T("SERVICE_CONTROL_INTERROGATE\n"));
			break; 
	} 
	if (!::SetServiceStatus(ServiceStatusHandle,  &ServiceStatus)) { 
		MyOutputDebug(_T("SetServiceStatus CtrlHandler()\n"));
		status = ::GetLastError(); 
	}
	return; 
}
//ms-help://MS.VSCC.v80/MS.MSDN.v80/MS.WIN32COM.v10.en/debug/base/adding_a_source_to_the_registry.htm
BOOL AddEventSource(
					LPTSTR pszLogName, // Application log or a custom log
					LPTSTR pszSrcName, // event source name
					LPTSTR pszMsgDLL,  // path for message DLL
					DWORD  dwNum)      // number of categories
{
	HKEY hk; 
	DWORD dwDisp; 
	TCHAR szBuf[MAX_PATH] = { 0 }; 
	// Create the event source as a subkey of the log. 
	lstrcat(szBuf, _T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\"));
	lstrcat(szBuf, pszLogName);
	lstrcat(szBuf, _T("\\"));
	lstrcat(szBuf, pszSrcName);
	long res = 0;
	res = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szBuf, 
		0, NULL, REG_OPTION_NON_VOLATILE,
		KEY_WRITE, NULL, &hk, &dwDisp);
	if(res) {
		//cout << _T("Could not create the registry key. Error code ") << GetLastError() << endl;
		return FALSE;
	}
	res = RegSetValueEx(hk,          // subkey handle 
		_T("EventMessageFile"),        // value name 
		0,                         // must be zero 
		REG_EXPAND_SZ,             // value type 
		(LPBYTE) pszMsgDLL,        // pointer to value data 
		(DWORD) lstrlen(pszMsgDLL)+1); // length of value data 
	if(res) {
		//cout << _T("Could not set the event message file. Error code ") << GetLastError() << endl;
		RegCloseKey(hk); 
		return FALSE;
	}
	RegCloseKey(hk); 
	return TRUE;
}
/*
	Install, remove this service dll.
	Use for install:
		regsvr32 /i syslog.dll
		for remove:
		regsvr32 /i /u syslog.dll
*/
extern "C" __declspec(dllexport) HRESULT DllRegisterServer(void)
{
	MyOutputDebug(_T("DllRegisterServer\n"));

	HRESULT hr = S_OK;
	SC_HANDLE hscm(0), hsvc(0);

	hscm = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	hsvc = ::CreateService(
		hscm,						// SCManager database 
		_T("SysLogService"),		// name of service 
		_T("SysLogService"),		// service name to display 
		SERVICE_ALL_ACCESS,			// desired access 
		SERVICE_WIN32_OWN_PROCESS,	// service type 
		SERVICE_AUTO_START,		// start type 
		SERVICE_ERROR_NORMAL,		// error control type 
		_T("%SystemRoot%\\system32\\svchost.exe -k syslogsvc"),					// service's binary 
		NULL,						// no load ordering group 
		NULL,						// no tag identifier 
		NULL,						// no dependencies 
		NULL,						// LocalSystem account 
		NULL);						// no password 

	if(hsvc == NULL) 
		hr = ::GetLastError();
	::CloseServiceHandle(hsvc);
	::CloseServiceHandle(hscm);
	if(hr == S_OK) {
		HKEY hk; 
		DWORD dwDisp; 
		TCHAR* czServValueParam = _T("SysLogService");
		TCHAR* czDescription = _T("          . RFC3164");

		HRESULT hr = ::RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
			_T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\SvcHost"),
			0,
			KEY_ALL_ACCESS,
			&hk);
		hr = ::RegSetValueEx(hk,					// subkey handle 
			_T("syslogsvc"),						// value name 
			0,										// must be zero 
			REG_MULTI_SZ,							// value type 
			(LPBYTE)czServValueParam,				// pointer to value data 
			(DWORD) lstrlen(czServValueParam)+1);	// length of value data 
		hr = ::RegCloseKey(hk); 
		hr = ::RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
			_T("SYSTEM\\CurrentControlSet\\Services\\SysLogService"),
			0,
			KEY_ALL_ACCESS,
			&hk);
		hr = ::RegSetValueEx(hk,					// subkey handle 
			_T("Description"),						// value name 
			0,										// must be zero 
			REG_EXPAND_SZ,							// value type 
			(LPBYTE)czDescription,					// pointer to value data 
			(DWORD) lstrlen(czDescription)+1);		// length of value data 
		hr = ::RegCloseKey(hk); 
		hr = ::RegCreateKeyEx(HKEY_LOCAL_MACHINE, _T("SYSTEM\\CurrentControlSet\\Services\\SysLogService\\Parameters"),
			0, NULL, REG_OPTION_NON_VOLATILE,
			KEY_WRITE, NULL, &hk, &dwDisp);
		hr = ::RegSetValueEx(hk,					// subkey handle 
			_T("ServiceDll"),						// value name 
			0,										// must be zero 
			REG_EXPAND_SZ,							// value type 
			(LPBYTE)czGModuleFilename,				// pointer to value data 
			(DWORD) lstrlen(czGModuleFilename)+1);	// length of value data 
		hr = ::RegCloseKey(hk);
		AddEventSource(_T("SysLog(UDP 514)"), _T("SysLogUdp514"), czGModuleFilename, 3);
	}
	::Sleep(10);
	return hr;
}
extern "C" __declspec(dllexport) HRESULT DllUnregisterServer(void)
{
	MyOutputDebug(_T("DllUnregisterServer\n"));
	HRESULT hr = S_OK;
	SC_HANDLE hscm(0), hsvc(0);
	HKEY hk; 

	hscm = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	hsvc = ::OpenService(hscm, _T("SysLogService"), SC_MANAGER_ALL_ACCESS);
	if(!::DeleteService(hsvc))
        hr = ::GetLastError();
	::CloseServiceHandle(hsvc);
	::CloseServiceHandle(hscm);
	if(!hr) {
		hr = ::RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
            _T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\SvcHost"),
            0,
            KEY_ALL_ACCESS,
            &hk);
        hr = ::RegDeleteValue(hk, _T("syslogsvc"));
        hr = ::RegCloseKey(hk); 
	}
	::Sleep(10);
	return hr;
}
