// syslog514.h
#pragma once
#include "iphlpadv.h"
#include "eventlog.h"
using namespace std;

class CSysLog
{
public:
	enum eSysLogState {
		esWorking = 0,
		esPaused,
		esStopped,
		esError
	};
	enum eSysLogCommand {
		ecPause = 0,
		ecResume,
		ecShutdown
	};
private:
	eSysLogState	mState;
	eSysLogCommand 	mCommand;
	HANDLE			mhThread;
	HANDLE			mhEventCommand; 
	WSADATA			mwsaData;
	int				mWSAErrorCode;
	TCHAR*			mpNameService;
	CEvent2Log		mLog;

public:
	CSysLog()
	{
		mState = CSysLog::esStopped;
		mCommand = CSysLog::ecResume;
		mhThread = 0;
		mWSAErrorCode = 0;
		memset(&mwsaData, 0, sizeof(mwsaData));
		mhEventCommand = ::CreateEvent(NULL, FALSE, FALSE, NULL);
		mpNameService = NULL;
		mLog.Report(_T(" "));
		return;
	}
	~CSysLog()
	{
		Shutdown();
		WSAStop();
		::CloseHandle(mhEventCommand);
		if(mpNameService) {
			/*
			 ,   .    ,  ,
			,    . , ,  
			  . .
			*/
			StartServiceByName(mpNameService);
			mLog.Report(_T("  , "));
		}
		mLog.Report(_T(" "));
		return;
	}
	eSysLogState State()
	{
		return mState;
	}
	void Enjoy()
	{
		mCommand = CSysLog::ecResume;
		if(!mhThread)
			mhThread = (HANDLE) _beginthread(CSysLog::SysLogThread, 0, this);
		Sleep(100);
		if(mState == CSysLog::esError && mWSAErrorCode == WSANOTINITIALISED) {
			mLog.Report(_T(":    "));
			Shutdown();
			WSAStart();
			mCommand = CSysLog::ecResume;
			Enjoy();
			return;
		}
		if(mState == CSysLog::esError && mWSAErrorCode == WSAEADDRINUSE) {
			mLog.Report(_T(":    "));
			Shutdown();
			KillProcess();
			mCommand = CSysLog::ecResume;
			Enjoy();
			return;
		}
		return;
	}
	void Shutdown()
	{
		if(mhThread) {
			mCommand = CSysLog::ecShutdown;
			::SetEvent(mhEventCommand);
			::WaitForSingleObject(mhThread, INFINITE);
			//CloseHandle(mhThread);
			mhThread = NULL;
			mLog.Report(_T(" "));
		}
		return;
	}
	void Pause()
	{
		mCommand = CSysLog::ecPause;
		::SetEvent(mhEventCommand);
		::Sleep(0);
		mLog.Report(_T(" "));
		return;
	}
	void Resume()
	{
		mCommand = CSysLog::ecResume;
		::SetEvent(mhEventCommand);
		::Sleep(0);
		mLog.Report(_T(" "));
		return;
	}
protected:
	BOOL WSAStart()
	{
		if(!mwsaData.wVersion) {
			WORD wVersionRequested = MAKEWORD(2, 2); 
			if(WSAStartup(wVersionRequested , &mwsaData)) {
				mLog.Report(_T("    "));
				mwsaData.wVersion = 0;
				WSACleanup();
				return FALSE;
			}
			mLog.Report(_T("   "));
		}
		return TRUE;
	}
	void WSAStop()
	{
		if(mwsaData.wVersion) WSACleanup();
		return;
	}
	void KillProcess()
	{
		DWORD pid = SearchPidPortListenerUDP(514);
		if(pid) {
			mLog.Report(_T("    "), EVENTLOG_WARNING_TYPE);
			mpNameService = StopServiceByPID(pid);
			if(!*mpNameService) {
				KillProcessByID(pid);
				mLog.Report(_T("   "), EVENTLOG_SUCCESS);
			} else {
				mLog.Report(_T("   () "), EVENTLOG_SUCCESS);
			}
		}
		return;
	}

private:
	static void __cdecl SysLogThread(LPVOID lpParam = NULL)
	{
		int			iRes;
		DWORD		dwRes;
		SOCKET		udpsocket;
		SOCKADDR_IN saI = { AF_INET, htons(514), INADDR_ANY };
		SOCKADDR_IN saO;
		char		c_buffer[1024];
		CSysLog&	obj = *(CSysLog*)lpParam;
		int			adrsize = sizeof(saO);
		string		str;
		stringstream ss;

		WSAEVENT evRead = WSACreateEvent();
		if(WSA_INVALID_EVENT == evRead) {
			iRes = -1;
			goto EXIT;
		}
		HANDLE		hevent[] = {obj.mhEventCommand, evRead};
		udpsocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if(INVALID_SOCKET == udpsocket) {
			iRes = -1;
			goto EXIT;
		}
		iRes = bind(udpsocket, (LPSOCKADDR)&saI, sizeof(saI));
		if(SOCKET_ERROR == iRes) goto EXIT;
		iRes = WSAEventSelect(udpsocket, evRead, FD_READ);
		if(SOCKET_ERROR == iRes) goto EXIT;
		obj.mState = CSysLog::esWorking;
		for(;;) {
			dwRes = WSAWaitForMultipleEvents(2, hevent, FALSE, WSA_INFINITE, FALSE);
			if(WSA_WAIT_EVENT_0 == dwRes) {
				switch(obj.mCommand) {
					case CSysLog::ecShutdown:
						closesocket(udpsocket);
						WSACloseEvent(evRead);
						obj.mLog.Report(_T("  stop"));
						obj.mState = CSysLog::esStopped;
						goto EXIT;
					case CSysLog::ecPause:
						obj.mLog.Report(_T("  pause"));
						obj.mState = CSysLog::esPaused;
						break;
					case CSysLog::ecResume:
						obj.mLog.Report(_T("  continue"));
						obj.mState = CSysLog::esWorking;
						break;
					default:
						obj.mState = CSysLog::esWorking;
				}
			} 
			if(WSA_WAIT_EVENT_0+1 == dwRes) {
				iRes = WSAResetEvent(evRead);
				if(!iRes) break;
				if(obj.mState == CSysLog::esWorking) {
					iRes = recvfrom(udpsocket, c_buffer, sizeof(c_buffer), 0, (LPSOCKADDR)&saO, &adrsize);
					if(iRes == SOCKET_ERROR) break;
					in_addr* a = &saO.sin_addr;
					c_buffer[iRes] = 0;
					WORD svr(0);
					ss << (int)a->S_un.S_un_b.s_b1 << '.'
						<< (int)a->S_un.S_un_b.s_b2 << '.'
						<< (int)a->S_un.S_un_b.s_b3 << '.'
						<< (int)a->S_un.S_un_b.s_b4 << ' '
						<< obj.BuildFieldPRI(c_buffer, &svr) << endl << ends;
					str = ss.str();
					ss.seekp(0, ios_base::beg );   
					obj.mLog.Report((char*)str.c_str(), svr);
				} else continue;
			}
		}
EXIT:
		if(iRes) {
			obj.mWSAErrorCode = WSAGetLastError();
			obj.mState = CSysLog::esError;
		}
		return;
	}
	char* BuildFieldPRI(const char* p, WORD* pw )
	{
		static char czRetBuffer[1024+512] = { 0 };
		static char* paFacility[] = { 
			"kernel",
			"user-level",
			"mail system",
			"system daemons",
			"security/authorization",
			"syslogd",
			"line printer subsystem",
			"network news subsystem",
			"UUCP subsystem",
			"clock daemon",
			"security/authorization messages",
			"FTP daemon",
			"NTP subsystem",
			"log audit",
			"log alert",
			"clock daemon",
			"local use 0",
			"local use 1",
			"local use 2",
			"local use 3",
			"local use 4",
			"local use 5",
			"local use 6",
			"local use 7",
			0
		};
		static char* paSeverity[] = { 
			"Emergency:",
			"Alert:",
			"Critical:",
			"Error:",
			"Warning:",
			"Notice:",
			"Informational:",
			"Debug:",
			0
		};
		//Read PRI code
		istringstream is;
		int iFacility = 0;
		int iSeverity = 0;
		int iPRIvalue = 0;
		string msg, s(p);
		basic_string <char>::size_type indexCh1a, indexCh1b;

		indexCh1a = s.find("<");
		indexCh1b = s.find(">");
		string b = s.substr(indexCh1a+1, indexCh1b-indexCh1a-1);

		is.str(b.c_str());
		is >> iPRIvalue;
		msg = s.substr(indexCh1b+1);

		iSeverity = iPRIvalue % 8;
		iFacility = (iPRIvalue - (iPRIvalue % 8))/8;

		czRetBuffer[0] = NULL;
		strcat(czRetBuffer, paSeverity[iSeverity]);
		strcat(czRetBuffer, paFacility[iFacility]);
		strcat(czRetBuffer, " ");
		strcat(czRetBuffer, msg.c_str());
		if(pw) {
			switch(iSeverity) {
				case 0: case 1: case 2: case 3:
					*pw = EVENTLOG_ERROR_TYPE;
					break;
				case 4:
					*pw = EVENTLOG_WARNING_TYPE;
					break;
				case 5:
					*pw = EVENTLOG_SUCCESS;
					break;
				default:
					*pw = EVENTLOG_INFORMATION_TYPE;
			}
		}
		/*
		PRI, HEADER, MSG
		int iFacility = 20;
		int iSeverity = 6;
		int iPRIvalue = 8 * iFacility + iSeverity;

		iSeverity = iPRIvalue % 8;
		iFacility = (iPRIvalue - (iPRIvalue % 8))/8;
		
		//EVENTLOG_SUCCESS                0x0000
		//EVENTLOG_ERROR_TYPE             0x0001
		//EVENTLOG_WARNING_TYPE           0x0002
		//EVENTLOG_INFORMATION_TYPE       0x0004
		//EVENTLOG_AUDIT_SUCCESS          0x0008
		//EVENTLOG_AUDIT_FAILURE          0x0010
		*/
		return czRetBuffer;
	}
};
